--
--	Sweeper Brush
--
--	@author: 	 Wopster
--	@descripion: Clear tracks and deform brush
--	@history:	 v1.0 - 2017-02-03 - Initial implementation
--

SweeperBrush = {}

function SweeperBrush.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(WorkArea, specializations) and SpecializationUtil.hasSpecialization(TurnOnVehicle, specializations) and SpecializationUtil.hasSpecialization(Shovel, specializations) and SpecializationUtil.hasSpecialization(DynamicHose, specializations)
end

function SweeperBrush:preLoad(savegame)	
    self.loadGroundParticleMapping = Utils.overwrittenFunction(self.loadGroundParticleMapping, SweeperBrush.loadGroundParticleMapping)
	self.loadEraseAreaFromXML = SweeperBrush.loadEraseAreaFromXML
	self.eraseTrack = SweeperBrush.eraseTrack
	self.getEraseAreas = SweeperBrush.getEraseAreas
	self.deformBrush = SweeperBrush.deformBrush
	self.setSweeperBrushHeight = SweeperBrush.setSweeperBrushHeight
	self.doCheckSpeedLimit = Utils.overwrittenFunction(self.doCheckSpeedLimit, SweeperBrush.doCheckSpeedLimit)	
	self.getIsTurnedOnAllowed = Utils.overwrittenFunction(self.getIsTurnedOnAllowed, SweeperBrush.getIsTurnedOnAllowed)
end

function SweeperBrush:load(savegame)	
	self.allowFillType = Utils.overwrittenFunction(self.allowFillType, SweeperBrush.allowFillType)
	self.allowUnitFillType = Utils.overwrittenFunction(self.allowUnitFillType, SweeperBrush.allowUnitFillType)
	self.setUnitFillLevel = Utils.overwrittenFunction(self.setUnitFillLevel, SweeperBrush.setUnitFillLevel)
	
	self.eraseAreas = {}
	
	local i = 0	
    while true do
        local key = ('vehicle.sweeperBrush.eraseAreas.eraseArea(%d)'):format(i)
       
        if not hasXMLProperty(self.xmlFile, key) then
            break
        end
		
        i = i + 1
		
		self:loadEraseAreaFromXML(self.xmlFile, key)
    end
	
	if self.isClient then
		local sampleNode = Utils.indexToObject(self.components, Utils.getNoNil(getXMLString(self.xmlFile, 'vehicle.sweeperBrush.brush.sound#linkNode'), '0>'))
		self.sampleBrush = SoundUtil.loadSample(self.xmlFile, {}, 'vehicle.sweeperBrush.brush.sound', '$data/maps/sounds/siloFillSound.wav', self.baseDirectory, sampleNode)
	end
	
	self.brushTurnedOnRotationNodes = Utils.loadRotationNodes(self.xmlFile, {}, 'vehicle.turnedOnRotationNodes.turnedOnRotationNode', 'sweeperBrush', self.components)
	
	self.brush = {
		node = Utils.indexToObject(self.components, getXMLString(self.xmlFile, 'vehicle.sweeperBrush.brush#index')),
		maxDeformation = Utils.getNoNil(getXMLFloat(self.xmlFile, 'vehicle.sweeperBrush.brush#maxDeformation'), 0),
		initialCompression = getXMLFloat(self.xmlFile, 'vehicle.sweeperBrush.brush#initialCompression'),
		yTravel = getXMLFloat(self.xmlFile, 'vehicle.sweeperBrush.brush#yTravel'),
		threshold = getXMLFloat(self.xmlFile, 'vehicle.sweeperBrush.brush#threshold'),
		shaderNodes = {},
		lastDeformation = 0
	}
	
	if self.brush.node ~= nil then
		-- load all child nodes that hold the shader
		for i = 0, getNumOfChildren(self.brush.node) - 1 do
			local shaderNode = getChildAt(self.brush.node, i)
			
			if shaderNode ~= nil and getHasShaderParameter(shaderNode, 'morphPosition') then
				if self.brush.x == nil or self.brush.y == nil or self.brush.z == nil then
					self.brush.x, self.brush.y, self.brush.z, _ = getShaderParameter(shaderNode, 'morphPosition')
				end
				
				setShaderParameter(shaderNode, 'morphPosition', self.brush.x, self.brush.y, self.brush.z, 0, false)
				
				table.insert(self.brush.shaderNodes, shaderNode)
			end
		end
		
		self.brush.deltaY = (1 - self.brush.initialCompression * 0.01) * self.brush.yTravel
	end
	
	self.heightAdjustment = {
		animation = getXMLString(self.xmlFile, 'vehicle.sweeperBrush.heightAdjustment#animationName'),
		inputBindingUp = InputBinding.higherHeight,
		inputBindingDown = InputBinding.lowerHeight,
		playerDistance = math.huge,
		playerMinDistance = 3,
		state = 0,
		inverse = false
	}
	
	self.heightAdjustmentEnabled = self.heightAdjustment.animation ~= nil
	self.isSweeperBrushSpeedLimitActive = false
	
	if self.configurations['brushColor'] ~= nil then
	    self:setColor(self.xmlFile, 'brushColor', self.configurations['brushColor'])
    end
end

function SweeperBrush:postLoad(savegame)
	if self.heightAdjustmentEnabled then
		if savegame ~= nil and not savegame.resetVehicles then
			local state = getXMLInt(savegame.xmlFile, savegame.key .. '#heightAdjustmentState')
			local inverse = getXMLBool(savegame.xmlFile, savegame.key .. '#heightAdjustmentInverse')
			
			self:setSweeperBrushHeight(state, inverse, true)
		end
	end
end

---
-- @param xmlFile
-- @param key
--
function SweeperBrush:loadEraseAreaFromXML(xmlFile, key)
	local start = Utils.indexToObject(self.components, getXMLString(self.xmlFile, key .. '#startIndex'))
    local width = Utils.indexToObject(self.components, getXMLString(self.xmlFile, key .. '#widthIndex'))
    local height = Utils.indexToObject(self.components, getXMLString(self.xmlFile, key .. '#heightIndex'))
	
	assert(start ~= nil and width ~= nil and height ~= nil, ('SweeperBrush: invalid area nodes! Please make sure the start, width, height index exist (%s)'):format(key))

	local groundReferenceIndex = getXMLInt(self.xmlFile, key .. '#refNodeIndex')
	
	assert(groundReferenceIndex ~= nil and self.groundReferenceNodes[groundReferenceIndex + 1] ~= nil, ('SweeperBrush: invalid ground reference node (%s)! Indexing starts with 0! (%s)'):format(groundReferenceIndex, key))

	table.insert(self.eraseAreas, {		
		start = start,
		width = width,
		height = height,
		refNode = self.groundReferenceNodes[groundReferenceIndex + 1]
	})
end

function SweeperBrush:delete()
end

function SweeperBrush:readStream(streamId, connection)
	if self.heightAdjustmentEnabled then
		self:setSweeperBrushHeight(streamReadUInt8(streamId), streamReadBool(streamId), true)
	end
end

function SweeperBrush:writeStream(streamId, connection)
	if self.heightAdjustmentEnabled then
		streamWriteUInt8(streamId, self.heightAdjustment.state)
		streamWriteBool(streamId, self.heightAdjustment.inverse)
	end
end

function SweeperBrush:getSaveAttributesAndNodes(nodeIdent)	
	if self.heightAdjustmentEnabled then
		return ('heightAdjustmentState="%s" heightAdjustmentInverse="%s"'):format(self.heightAdjustment.state, self.heightAdjustment.inverse)
	end
	
    return nil
end

function SweeperBrush:mouseEvent(posX, posY, isDown, isUp, button)
end

function SweeperBrush:keyEvent(unicode, sym, modifier, isDown)
end

function SweeperBrush:update(dt)
	if self.isClient then
		Utils.updateRotationNodes(self, self.brushTurnedOnRotationNodes, dt, self:getIsActive() and self:getIsTurnedOn())
    end
	
	-- deformation of brush
	if self.brush.node ~= nil then
		self:deformBrush()		
	end
	
	if self.heightAdjustmentEnabled and SweeperBrush:isPlayer() then
		self.heightAdjustment.playerDistance = math.huge
		
		local playerTrans = {getWorldTranslation(g_currentMission.player.rootNode)}
		local toolTrans = {getWorldTranslation(self.rootNode)}
		local distance = Utils.vector3Length(playerTrans[1] - toolTrans[1], playerTrans[2] - toolTrans[2], playerTrans[3] - toolTrans[3])
		
		if distance < self.heightAdjustment.playerDistance and distance < self.heightAdjustment.playerMinDistance then
			local alpha = nil
			local inverse = false
						
			if InputBinding.isPressed(self.heightAdjustment.inputBindingDown) and not InputBinding.isPressed(self.heightAdjustment.inputBindingUp) then
				-- lower
				alpha = Utils.clamp(self.heightAdjustment.state + (dt / 2000), 0, 1)
			elseif InputBinding.isPressed(self.heightAdjustment.inputBindingUp) and not InputBinding.isPressed(self.heightAdjustment.inputBindingDown) then
				-- heigher
				alpha = Utils.clamp(self.heightAdjustment.state - (dt / 2000), 0, 1)
				inverse = true
			end
						
			if alpha ~= nil and alpha ~= self.heightAdjustment.state then
				self:setSweeperBrushHeight(alpha, inverse)
			end
			
			if self.heightAdjustment.state < 1 and self.heightAdjustment.state >= 0 then
				g_currentMission:addHelpButtonText(g_i18n:getText('input_lowerHeight'), self.heightAdjustment.inputBindingDown)		
			end

			if self.heightAdjustment.state <= 1 and self.heightAdjustment.state > 0 then
				g_currentMission:addHelpButtonText(g_i18n:getText('input_higherHeight'), self.heightAdjustment.inputBindingUp)		
			end
			
			self.heightAdjustment.playerDistance = distance
		end
	end
end

function SweeperBrush:updateTick(dt)
	self:eraseTrack(dt)
end

function SweeperBrush:draw()
	-- inform about more function outside vehicle?
	if self.heightAdjustmentEnabled then
		g_currentMission:addExtraPrintText(g_i18n:getText('info_heightAdjustment'))
	end
end

---
-- @param dt
--
function SweeperBrush:eraseTrack(dt)
	if g_currentMission.tireTrackSystem ~= nil then
		self.isSweeperBrushSpeedLimitActive = false
		
		local eraseAreas = self:getEraseAreas() -- clear everything or only on object contact?
		
		if self:getIsActive() then		
			if #eraseAreas > 0 then
				for _, eraseArea in pairs(eraseAreas) do
					g_currentMission.tireTrackSystem:eraseParallelogram(eraseArea[1], eraseArea[2], eraseArea[3], eraseArea[4], eraseArea[5], eraseArea[6])
				end
				
				self.isSweeperBrushSpeedLimitActive = true
				
				if self.isServer then
					local fillLevel = self:getUnitFillLevel(self.shovel.fillUnitIndex)
					local capacity = self:getUnitCapacity(self.shovel.fillUnitIndex)
					
					if fillLevel < capacity then						
						if self:allowFillType(FillUtil.FILLTYPE_MANURE) then
							local deltaFill = self.shovelFillLitersPerSecond * dt * 0.001 / 50 -- fill it slowly
							
							if deltaFill > 0 then -- always overwrite the fillType cause we are cleaning right?
								self:setUnitFillLevel(self.shovel.fillUnitIndex, fillLevel + deltaFill, FillUtil.FILLTYPE_MANURE, true, self.fillVolumeLoadInfos[self.shovel.unloadInfoIndex])
							end
						end
					end
				end
				
				if self.sweeperBrushColorParticleSystems ~= nil then
					for _, mapping in pairs(self.sweeperBrushColorParticleSystems) do
						local trans = {getWorldTranslation(mapping.node)}
						local densityBits = getDensityAtWorldPos(g_currentMission.terrainDetailId, trans[1], trans[2], trans[3])
						
						if densityBits ~= 0 then
							mapping.lastColor[1], mapping.lastColor[2], mapping.lastColor[3], _ = Utils.getTireTrackColorFromDensityBits(densityBits)
						else
							mapping.lastColor[1], mapping.lastColor[2], mapping.lastColor[3], _, _ = getTerrainAttributesAtWorldPos(g_currentMission.terrainRootNode, trans[1], trans[2], trans[3], true, true, true, true, false)
						end
						
						-- interpolate between different ground colors to avoid unrealisitic particle color changes
						if mapping.targetColor == nil then
							mapping.targetColor = mapping.lastColor
							mapping.currentColor = mapping.lastColor
							mapping.alpha = 1
						end

						if mapping.alpha ~= 1 then
							mapping.alpha = math.min(mapping.alpha + dt / 1000, 1)
							mapping.currentColor = {Utils.vector3ArrayLerp(mapping.lastColor, mapping.targetColor, mapping.alpha)}
							
							if mapping.alpha == 1 then
								mapping.lastColor = mapping.currentColor
							end
						end
						
						if mapping.alpha == 1 and mapping.lastColor[1] ~= mapping.targetColor[1] and mapping.lastColor[2] ~= mapping.targetColor[2] and mapping.lastColor[3] ~= mapping.targetColor[3] then
							mapping.targetColor = mapping.lastColor
							mapping.alpha = 0
						end
		
						setShaderParameter(mapping.particleSystem.shape, 'psColor', mapping.currentColor[1], mapping.currentColor[2], mapping.currentColor[3], 1, false)
					end
				end
			end
		end
		
		if self.isClient then
			if self:getIsActive() and #eraseAreas > 0 and self:getIsTurnedOn() then
				if self:getIsActiveForSound() then
					SoundUtil.playSample(self.sampleBrush, 0, 0, nil)
					SoundUtil.stop3DSample(self.sampleBrush)
				else					
					SoundUtil.stopSample(self.sampleBrush)
					SoundUtil.play3DSample(self.sampleBrush)
				end			
			else
				SoundUtil.stopSample(self.sampleBrush)
				SoundUtil.stop3DSample(self.sampleBrush)
			end
		end		
    end
end

---
-- @param eraseArea
--
function SweeperBrush:getIsEraseAreaActive(eraseArea)
	return eraseArea.refNode ~= nil and eraseArea.refNode.isActive	
end

---
--
function SweeperBrush:getEraseAreas()
	local activeAreas = {}
	
	if self:getIsTurnedOn() then
		for _, eraseArea in pairs(self.eraseAreas) do
			if SweeperBrush:getIsEraseAreaActive(eraseArea) then
				local start = {getWorldTranslation(eraseArea.start)}
				local width = {getWorldTranslation(eraseArea.width)}
				local height = {getWorldTranslation(eraseArea.height)}
				
				table.insert(activeAreas, {start[1], start[3], width[1], width[3], height[1], height[3]})
			end
		end
	end
	
	return activeAreas
end

---
-- @param superFunc
-- @param xmlFile
-- @param key
-- @param mapping
-- @param index
-- @param i3dNode
--
function SweeperBrush:loadGroundParticleMapping(superFunc, xmlFile, key, mapping, index, i3dNode)
    if superFunc ~= nil then
        if not superFunc(self, xmlFile, key, mapping, index, i3dNode) then
            return false
        end
    end

    mapping.adjustColor = Utils.getNoNil(getXMLBool(xmlFile, key .. '#adjustColor'), false)
	
    if mapping.adjustColor then
        if self.sweeperBrushColorParticleSystems == nil then
            self.sweeperBrushColorParticleSystems = {}
        end
		
        mapping.lastColor = {}

        table.insert(self.sweeperBrushColorParticleSystems, mapping)
    end

    return true
end

---
-- @param superFunc
-- @param isTurnedOn
--
function SweeperBrush:getIsTurnedOnAllowed(superFunc, isTurnedOn)
	if not self:getIsHoseAttached('hydraulic') then
		return false		
	end
	
	return superFunc(self, isTurnedOn)
end

---
-- @param superFunc
-- @param fillType
-- @param allowEmptying
--
function SweeperBrush:allowFillType(superFunc, fillType, allowEmptying)
	if self.isSweeperBrushSpeedLimitActive and self.fillUnits[self.shovel.fillUnitIndex].fillTypes[fillType] then
		return true
	end
	
	return superFunc(self, fillType, allowEmptying)
end

---
-- @param superFunc
-- @param fillUnitIndex
-- @param fillType
-- @param allowEmptying
--
function SweeperBrush:allowUnitFillType(superFunc, fillUnitIndex, fillType, allowEmptying)
	if self.isSweeperBrushSpeedLimitActive and self.fillUnits[fillUnitIndex].fillTypes[fillType] then
		return true
	end
	
	return superFunc(self, fillUnitIndex, fillType, allowEmptying)
end

---
-- @param superFunc
-- @param fillUnitIndex
-- @param fillLevel
-- @param fillType
-- @param force
-- @param fillInfo
--
function SweeperBrush:setUnitFillLevel(superFunc, fillUnitIndex, fillLevel, fillType, force, fillInfo)
	if self.isSweeperBrushSpeedLimitActive then
		force = true
	end
	
	if superFunc ~= nil then
		superFunc(self, fillUnitIndex, fillLevel, fillType, force, fillInfo)
	end
end

---
-- @return bool
--
function SweeperBrush:isPlayer()
	return g_currentMission.controlPlayer and g_currentMission.player ~= nil and g_gui.currentGui == nil and not g_currentMission.isPlayerFrozen and not g_currentMission.player.hasHPWLance and g_currentMission.player.currentTool == nil and not g_currentMission.player.isCarryingObject
end

---
-- @param superFunc
--
function SweeperBrush:doCheckSpeedLimit(superFunc)
	local parent = false
	
    if superFunc ~= nil then
        parent = superFunc(self)
    end

    return parent or self.isSweeperBrushSpeedLimitActive
end

---
--
function SweeperBrush.getDefaultSpeedLimit()
	return 10
end

---
--
function SweeperBrush:deformBrush()
	local x, y, z = getWorldTranslation(self.brush.node)
	local terrainHeight = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, x, y - self.brush.yTravel, z) -- only do on terrain contact for now..	
	local difference = (terrainHeight + self.brush.threshold) - y
	
	if math.abs(difference) < 1 then
		local deformation = Utils.clamp((self.brush.deltaY + difference) * 0.7, 0, self.brush.maxDeformation)
		
		if self.brush.lastDeformation ~= deformation then
			self.brush.lastDeformation = deformation
			 
			if #self.brush.shaderNodes > 0 then
				for _, shaderNode in pairs(self.brush.shaderNodes) do
					setShaderParameter(shaderNode, 'morphPosition', self.brush.x, self.brush.y, self.brush.z, self.brush.lastDeformation, false)	
				end
			end
		end
	end
end

---
-- @param state
-- @param noEventSend
--
function SweeperBrush:setSweeperBrushHeight(state, inverse, noEventSend)
	SweeperBrushHeightAdjustmentEvent.sendEvent(self, state, inverse, noEventSend)

	local multiplier = 10 ^ (2)
	state = math.floor(state * multiplier + 0.5) / multiplier
	local animTime = (state * self:getAnimationDuration(self.heightAdjustment.animation)) / self:getAnimationDuration(self.heightAdjustment.animation)
	
	-- always stop the animation to make sure the state is set correctly
	self:stopAnimation(self.heightAdjustment.animation, true)
	
	local direction = inverse and -1 or 1
	
	self:setAnimationStopTime(self.heightAdjustment.animation, animTime)
	self:playAnimation(self.heightAdjustment.animation, direction, self:getAnimationTime(self.heightAdjustment.animation), true)
	
	self.heightAdjustment.state = state
	self.heightAdjustment.inverse = inverse
	
	-- only inverse direction when clamped value is reached
	-- if state >= 1 or state <= 0 then
		-- self.heightAdjustment.inverse = state >= 1
	-- end
end

ConfigurationUtil.registerConfigurationType('brushColor', g_i18n:getText('configuration_brushColor'), nil, Vehicle.getConfigColorSingleItemLoad, Vehicle.getConfigColorPostLoad, ConfigurationUtil.SELECTOR_COLOR)

--
--	Sweeper Brush: height adjustment event
--

SweeperBrushHeightAdjustmentEvent = {}
SweeperBrushHeightAdjustmentEvent_mt = Class(SweeperBrushHeightAdjustmentEvent, Event)

InitEventClass(SweeperBrushHeightAdjustmentEvent, 'SweeperBrushHeightAdjustmentEvent')

function SweeperBrushHeightAdjustmentEvent:emptyNew()
    local self = Event:new(SweeperBrushHeightAdjustmentEvent_mt)

    return self
end

function SweeperBrushHeightAdjustmentEvent:new(object, state, inverse)
    local self = SweeperBrushHeightAdjustmentEvent:emptyNew()
    
	self.object = object
	self.state = state
	self.inverse = inverse

    return self
end

function SweeperBrushHeightAdjustmentEvent:writeStream(streamId, connection)
    writeNetworkNodeObject(streamId, self.object)
    streamWriteFloat32(streamId, self.state)
    streamWriteBool(streamId, self.inverse)
end

function SweeperBrushHeightAdjustmentEvent:readStream(streamId, connection)
    self.object = readNetworkNodeObject(streamId)
    self.state = streamReadFloat32(streamId)
    self.inverse = streamReadBool(streamId)
    self:run(connection)
end

function SweeperBrushHeightAdjustmentEvent:run(connection)
	if self.object ~= nil then
		self.object:setSweeperBrushHeight(self.state, self.inverse, true)
    end
	
	if not connection:getIsServer() then
	    g_server:broadcastEvent(SweeperBrushHeightAdjustmentEvent:new(self.object, self.state, self.inverse), nil, connection, self.object)
	end
end

function SweeperBrushHeightAdjustmentEvent.sendEvent(object, state, inverse, noEventSend)
    if noEventSend == nil or noEventSend == false then
        if g_server ~= nil then
            g_server:broadcastEvent(SweeperBrushHeightAdjustmentEvent:new(object, state, inverse), nil, nil, object)
        else
            g_client:getServerConnection():sendEvent(SweeperBrushHeightAdjustmentEvent:new(object, state, inverse))
        end
    end
end